--[[
Restoration Shaman Stat Analyzer, by Rheeah

Expanded and enhanced version based on the original "Resto Shaman Mastery Analyzer" addon by _swift
See:
http://www.curse.com/addons/wow/rshamanmastery

--]]

local revision = tonumber(string.sub("$Revision: 27 $", 12, -3))
_G.RShamanStat = {
    ["revision"] = revision
}
local rss = _G.RShamanStat

-- item id's for gems that increase critical effect; only MoP gems that could be sensibly used by healers are included
local BURNING_PRIMAL_DIAMOND = 76885
local REVITALIZING_PRIMAL_DIAMOND = 76888
local criticalBonusGems = { BURNING_PRIMAL_DIAMOND, REVITALIZING_PRIMAL_DIAMOND }

-- item id's for T15 tier pieces 
local setBonusT15 = {     ["ChestSlot"] = { 95940 , 95310 , 96684 } , 
						  ["HandsSlot"] = { 95941 , 95311 , 96685 } , 
						   ["HeadSlot"] = { 95942 , 95312 , 96686 } ,
						   ["LegsSlot"] = { 95943 , 95313 , 96687 } , 
					   ["ShoulderSlot"] = { 95944 , 95314 , 96688 } }

-- spell id's for various heals; some are not needed (yet) and some item proc's that heal should probably be added
local HEALING_WAVE = 331
local HEALING_SURGE = 8004
local GREATER_HEALING_WAVE = 77472
local RIPTIDE = 61295
local ANCESTRAL_AWAKENING = 51558
local UNLEASH_LIFE = 73685
local CHAIN_HEAL = 1064
local HEALING_RAIN = 73920
local EARTHLIVING = 51945
local EARTH_SHIELD = 974
local RESTORATIVE_MISTS = 114083         -- Ascendance
local RESURGENCE = 101033                -- Mana Gains are logged under combatEvent SPELL_ENERGIZE
local CLEANSING_WATERS = 55562			 -- heal from glyphed Purify Spirit
local SUMMON_HEALING_STREAM = 5394       -- the spell which summons HST
local SUMMON_HEALING_TIDE = 108280       -- the spell which summons HTT
local SUMMON_SPIRIT_LINK = 98008         -- the spell which summons SLT
local HEALING_STREAM = 52042             -- the actual heal applied by HST
local HEALING_STREAM_2T15 = 138304       -- the 'bonus' heal applied by HST with the T15 set bonus
local HEALING_TIDE = 114942              -- the actual heal applied by HTT
local SPIRIT_LINK = 98021          		 -- the actual heal applied by SLT
local PROC_RESTORATION_INFINITE = 139195 -- Lei Shen trinket proc

-- A table that links SpellIDs with their respective resurgence based mana gains
local RESURGENCE_GAINS = {
    [HEALING_WAVE] = 8849,
    [GREATER_HEALING_WAVE] = 8849,
    [HEALING_SURGE] = 8849 * 0.6,
    [RIPTIDE] = 8849 * 0.6,
    [UNLEASH_LIFE] = 8849 * 0.6,
    [CHAIN_HEAL] = 8849 * 0.333
}


local aaTriggers = { HEALING_SURGE, HEALING_WAVE, GREATER_HEALING_WAVE, RIPTIDE, UNLEASH_LIFE, CLEANSING_WATERS, PROC_RESTORATION_INFINITE }

local criticalMultiplier = 2.0           -- multiplier for critical effects; default value is 2
local aaSetBonusT15 = false              -- flag to indicate whether the 4T15 set bonus is applicable

local lastCast = { }                     -- table of most recent cast of each spell (indexed by spell name), for matching with AA
local healthTable = { }                  -- table of most recent health and max health of every known friendly unit
local spiritLinkGUID = nil               -- ID of our currently summoned Spirit Link totem, or nil

local fight = nil;


-- check for a crit-enhancing metagem
function RShamanStatCheckMeta()
	local gem1, gem2, gem3 = GetInventoryItemGems(GetInventorySlotInfo("HeadSlot")); -- head slot
	if tContains(criticalBonusGems, gem1) or tContains(criticalBonusGems, gem2) or tContains(criticalBonusGems, gem3) then
		criticalMultiplier = 2.0 * 1.03;
	else
		criticalMultiplier = 2.0;
	end
end

function RShamanStatCheckTier()
	aaSetBonusT15 = false;
	local n15 = 0;
	for slotName, tierItems in pairs(setBonusT15) do
		local equipID = GetInventoryItemID("player", GetInventorySlotInfo(slotName));
		if equipID and tContains(tierItems, equipID) then
			n15 = n15 + 1;
		end
	end
	if n15 >= 4 then
		aaSetBonusT15 = true;
	end
end

-- initialization
function RShamanStatOnLoad(self)
	local lClass, eClass = UnitClass("player");
	if eClass ~= "SHAMAN" then
		return;
	end
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	self:RegisterEvent("PLAYER_REGEN_ENABLED");
	self:RegisterEvent("PLAYER_REGEN_DISABLED");
	self:RegisterEvent("UNIT_MAXHEALTH");
	self:RegisterEvent("UNIT_HEALTH_FREQUENT");
	self:RegisterEvent("VARIABLES_LOADED");
	RShamanStatClear();
	print(("Resto Shaman Stat Analyzer (r%s) loaded!"):format(rss.revision));
end


function RShamanStatPrintRSM()
    if (#RShamanStatArchive == 0) then
        print("RSS: no data available");
        return;
    end

    for i, v in ipairs(RShamanStatArchive) do
        if (RShamanStatArchive.filter.useAll or v.isBoss) then
            if v.effectiveTotal ~= 0 then
                print(("RSS: #%02d - %s: Healing Done: %.1fk, Mastery Effect: %.2f"):format(i, v.encounterName, v.effectiveTotal/1000, v.weightedTotal/v.effectiveTotal));
            else
                print(("RSS: #%02d - %s: No Healing Done"):format(i, v.encounterName));
            end
        end
    end
end


-- erase current combat data AND all archived/saved data
function RShamanStatClearAll()
    RShamanStatClear();
    RShamanStatArchive = rss.CreateArchive();
end

-- erase current combat data; called when entering or leaving combat
function RShamanStatClear()
    lastCast = { };
    healthTable = { };
    spiritLinkGUID = nil;
    fight = rss.CreateRecord();

    RShamanStatCheckMeta();
    RShamanStatCheckTier();
end


-- Print data from combat that just ended
-- Save it to archive table
-- Zero accumulators for next combat
function RShamanStatReset()
	if fight.effectiveTotal > 0 then
        rss.SaveRecord(RShamanStatArchive, fight)
        rss.PrintRecord(fight)
    end
	RShamanStatClear()
end


SLASH_RSHAMANSTAT1 = '/rss';

function RShamanStatPrintDifficulty()
    if (RShamanStatArchive.filter.difficulty ~= 0) then
        local difficultyName = GetDifficultyInfo(RShamanStatArchive.filter.difficulty);
        print("Instance difficulty is set to '"..difficultyName.."'");
    else
        print("Instance difficulty is set to 'Any'");
    end
end

function RShamanStatPrintUseAll()
    if (RShamanStatArchive.filter.useAll) then
        print("Recording all encounters.");
    else
        print("Recording boss encounters only.");
    end
end

-- Supported commands:
function RShamanStatCmd(msg, ...)
    local msg, args = msg:match("^(%S*)%s*(.-)$");
    if (msg == "help" or msg == "") then
        print(("Resto Shaman Stats (r%s) - Available Commands:"):format(revision));
        print(" * /rss sum          Print all data, and grand totals");
        print(" * /rss sum <boss>   Print data for the specified boss");
        print(" * /rss clr            Print all data, and grand totals, then wipe all data");
        print(" * /rss boss          Only include boss encounter data for grand totals");
        print(" * /rss all            Include all data for grand totals");
        print(" * /rss rsm            Print data of last encounter in RSM mode");
        print(" * /rss difficulty <Any,10N,10H,25H,25N,LFR> Print data from a specific difficulty only");


        RShamanStatPrintUseAll()
        RShamanStatPrintDifficulty()
        return;
    end

    if (msg == "rsm") then
        RShamanStatPrintRSM();
        return;
    end

    if (msg == "sum" and args ~= "") then
        local r = rss.CreateAggregateRecord(args, RShamanStatArchive, function(r) return (r.encounterName == args); end)
        rss.args = args
        rss.PrintRecord(r)
        return;
    end

    if (msg == "difficulty" and args ~= "") then
        if args == "10N" then
            RShamanStatArchive.filter.difficulty = 3;
        elseif args == "25N" then
            RShamanStatArchive.filter.difficulty = 4;
        elseif args == "10H" then
            RShamanStatArchive.filter.difficulty = 5;
        elseif args == "25H" then
            RShamanStatArchive.filter.difficulty = 6;
        elseif args == "LFR" then
            RShamanStatArchive.filter.difficulty = 7;
        else
            RShamanStatArchive.filter.difficulty = 0;
        end
        RShamanStatPrintDifficulty()
        return;
    end

    if (msg == "boss") then
        RShamanStatArchive.filter.useAll = false;
        RShamanStatPrintUseAll()
        return
	elseif (msg == "all") then
        RShamanStatArchive.filter.useAll = true;
        RShamanStatPrintUseAll()
        return
    end



	if (msg == "sum" or msg == "clr") then
		local tot = 0;
		local timeTot = 0;
		local crtTot = 0;
		local mstTot = 0;
		local crtInc = 0;
		local crtIncAll = 0;
		local mstInc = 0;

        local filter = function(r)
            return (RShamanStatArchive.filter.useAll or r.isBoss) and (RShamanStatArchive.filter.difficulty == 0 or RShamanStatArchive.filter.difficulty == r.difficulty)
        end
		if (#RShamanStatArchive > 0) then
            for i, v in ipairs(RShamanStatArchive) do
                if (filter(v)) then
                    rss.PrintRecord(v)
                end
			end
        end

        local summary = rss.CreateAggregateRecord("Summary", RShamanStatArchive, filter)
        rss.PrintRecord(summary)
		print("---------------");
	end
	if (msg == "clr") then
		RShamanStatClearAll();
		print("RSS:: All archived data cleared");
	end
end
SlashCmdList["RSHAMANSTAT"] = RShamanStatCmd;

-- returns the name and classification of any hostile target of anyone we heal (boss detection)
function RShamanStatCheckTarget(unit)
	local unitTarget = unit.."-target";
	if ( UnitExists(unitTarget) and UnitAffectingCombat(unitTarget) and UnitCanAttack(unit, unitTarget) ) then
		return UnitName(unitTarget), UnitClassification(unitTarget);
	end
end

-- sets boss flag if worldboss target is detected, and sets encounter name
function RShamanStatCheckBoss(unit)
	local name, classification = RShamanStatCheckTarget(unit);
	if (classification == "worldboss" or (name ~= nil and fight.encounterName == "Unknown")) then
		fight.encounterName = name;
	end
	if (classification == "worldboss") then
		fight.isBoss = true;
		print("RSS:: Detected boss encounter "..name);
	end
end

function RShamanStatCombatLogEntry(...)
	local playerGUID = UnitGUID("player");
	local timestamp, combatEvent, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags = select(1, ...);
	
	if sourceGUID == playerGUID and combatEvent and strfind(combatEvent, 'SPELL_SUMMON') then
		local spellId, spellName, spellSchool = select(12, ...);
		if spellId == SUMMON_SPIRIT_LINK then 
			spiritLinkGUID = destGUID;
		end
		return;
	end
	
	if combatEvent and strfind(combatEvent, 'UNIT_DESTROYED') and destGUID == spiritLinkGUID then
		spiritLinkGUID = nil;
		return;
    end

    -- handle power gains like resurgence
    if sourceGUID == playerGUID and combatEvent and combatEvent == "SPELL_ENERGIZE" then
        local spellId, spellName, spellSchool = select(12, ...)
        local amount, powerType = select(15, ...)
        if spellId == RESURGENCE then
            fight.resurgence = fight.resurgence + amount
        end
        return;
    end

	
	local spiritLinkHeal = spiritLinkGUID and sourceGUID == spiritLinkGUID;
	if (sourceGUID == playerGUID or spiritLinkHeal) and combatEvent and strfind(combatEvent, '_HEAL') and healthTable[destGUID] ~= nil then
		-- Check for boss encounter
		if ( (not fight.isBoss) or (fight.encounterName == "Unknown") ) then
			RShamanStatCheckBoss(destName);
		end
		-- Check for first action of combat
		if fight.startTime <= 0 then
            fight.startTime = timestamp;
		end
        fight.totalTime = timestamp - fight.startTime;
		
		-- We healed somebody
		local spellId, spellName, spellSchool = select(12, ...)
		local amount, overhealing, absorbed, critical = select(15, ...)
		local healed = amount - overhealing;                                    -- ignore overhealing
        fight.effectiveTotal = fight.effectiveTotal + healed;                               -- accumulate total effective throughput
		
		if spiritLinkHeal then 													-- stat effects on Spirit Link totem heals doubtful
			return;
        end

        -- count the number of times we healed with non-periodic spells
        if combatEvent == "SPELL_HEAL" then
            local spellId_count = fight.spells[spellId] or 0
            spellId_count = spellId_count + 1
            fight.spells[spellId] = spellId_count
        end

		
		-- Deal with Mastery
		local myMasteryBonus = 0.03*GetMastery();  								-- this is the % value show on character panel, including Grace of Air
		local unitHealthTable = healthTable[destGUID];							-- find last known values of health and max health
		local hpratio = 1 - (unitHealthTable.currentHealth/unitHealthTable.maxHealth);
		local healBaseAmount = amount/(1+hpratio*myMasteryBonus);               -- the size of the heal with 0% Mastery
		local masteryGain = amount - healBaseAmount;                            -- increase from base amount from Mastery
		local effectiveMasteryGain = math.max(0.0, masteryGain - overhealing);  -- overhealing is not useful
		local incrementalMasteryGain = 0.0;
		if (overhealing == 0) then
			incrementalMasteryGain = 0.00005 * hpratio * healBaseAmount;        -- how much would +1 Mastery Rating increase this heal? (no benefit if already overhealing)
        end
        fight.weightedTotal = fight.weightedTotal + healed * hpratio
        fight.masteryTotal = fight.masteryTotal + effectiveMasteryGain;						-- accumulate gain from current Mastery
        fight.incrementalMastery = fight.incrementalMastery + incrementalMasteryGain;       -- accumulate gain from +1 Mastery
		
		-- Deal with Crit
        local myCriticalChance = GetSpellCritChance(log(spellSchool+1) / log(2))/100;  -- spell critical chance, spellSchool is determined using log2(spellSchool)
		local thisCast = nil;                                                   -- record this as the most recent cast of this spell, for matching with AA
		if spellId ~= ANCESTRAL_AWAKENING then                                  -- unless it is AA healing, which can't trigger AA
			thisCast = { };
			thisCast.amount = amount
			thisCast.critChance = myCriticalChance;
			thisCast.critical = false;
		end
		if (critical or (spellId == ANCESTRAL_AWAKENING)) then                                -- special logic for AA or Tidal Waves -> Healing Surge
			if spellId == HEALING_SURGE then                                           -- is this a Healing Surge critical?
				if UnitAura("player", "Tidal Waves") ~= nil then              -- tested: even if this cast consumed the last TW stack, we still have the buff at this moment
					myCriticalChance = math.min(1.0, myCriticalChance + 0.3);   -- adjust HS critical chance for TW
					thisCast.critChance = myCriticalChance;                     -- record the buffed value of critical chance for use with any AA triggered by this heal
				end
			end
			if (critical) then                                                  -- this cast crit, so figure out how much we gained over the base heal
				thisCast.critical = true;                                       -- record the cast as critical so we can add the triggered AA to the critical category
				local critBaseAmount = amount/criticalMultiplier;               
				local critBonusAmount = amount - critBaseAmount;
                fight.criticalTotal = fight.criticalTotal + math.max(0.0, critBonusAmount - overhealing);  -- record the effective gain from this crit
				-- how much would +1 Critical Rating increase the *expected* (average) effective amount of this cast?
				if (myCriticalChance > 0) then
                    fight.incrementalCritical = fight.incrementalCritical + (0.01/600)*math.max(0.0, critBonusAmount - overhealing)/myCriticalChance;
				end
			else -- Ancestral Awakening                                      
				local mindiff = amount;                                         -- Match this AA to the spell that triggered it, based on the 30% relationship
				local sMin = nil;
				local vMin = nil;
				for s, v in pairs(lastCast) do
					if math.abs(v.amount*0.3 - amount) < mindiff then
						mindiff = math.abs(v.amount*0.3 - amount);
						vMin = v;
						sMin = s;
					end
				end
				if vMin ~= nil then
					if vMin.critical then                          				   -- if the trigger for this AA was a crit (rather than the set bonus)..
					    local nonCritAA = 0;                                       -- size of a possible non-crit AA from original heal; zero unless set bonus present
						if aaSetBonusT15 then
							nonCritAA = 0.5 * 0.3 * vmin.amount / criticalMultiplier     -- with set bonus, half of 30% of the non-crit heal amount expected
						end
                        fight.criticalTotal = fight.criticalTotal + math.max(0, healed - nonCritAA); -- count it as gain from critical
						if (vMin.critChance > 0) then
							-- how much would +1 Critical Rating increase the expected (average) value of the cast that triggered it?
                            fight.incrementalCritical = fight.incrementalCritical + (0.01/600)*math.max(0, healed - nonCritAA)/vMin.critChance;
						end
					else  -- this AA WAS triggered by the set bonus; remove the current (non-crit) AA amount from the gain we added previously
                        fight.incrementalCriticalAll = math.max(0, fight.incrementalCriticalAll - (0.01/600)*healed);
					end
				end
			end
		else -- this is not a critical or an AA
			local criticalAmount = amount * criticalMultiplier;                -- how big would the raw heal be if it crit?
			local criticalAA = 0.3 * criticalAmount;                           -- how big would the AA heal be if it crit?
			if overhealing == 0 then										   -- no benefit to the target from critting on an overheal
				local criticalHeal = math.min(criticalAmount, unitHealthTable.maxHealth - unitHealthTable.currentHealth);
                fight.incrementalCriticalAll = fight.incrementalCriticalAll + (0.01/600)*(criticalHeal - healed);
			end
			-- add healing from the hypothetical AA, if applicable; we assume it would hit the friendly on our list with the largest health deficit ("most injured")
			if tContains(aaTriggers, spellId) and (spellId ~= RIPTIDE or strfind(combatEvent, 'PERIODIC_HEAL') == nil) then
				local maxDeficit = 0;
				for guid, t in pairs(healthTable) do
					local deficit = math.max(0, t.maxHealth - t.currentHealth);
					if guid == destGUID then                                  -- account for the amount we *would* have healed the target...
						deficit = math.max(0, deficit - criticalAmount);
					end
					maxDeficit = math.max(maxDeficit, deficit);
				end
				criticalAA = math.min(criticalAA, maxDeficit);
                fight.incrementalCriticalAll = fight.incrementalCriticalAll + (0.01/600)*criticalAA;
			end
		end
		if (thisCast ~= nil) then                                              -- store information on this cast for AA matching and accounting
			lastCast[spellName] = thisCast;
		end
	-- done with Mastery and Crit
	end
end
	
function RShamanStatOnEvent(self, event, ...)
	local currentSpec = GetSpecialization()
	local role = currentSpec and select(6, GetSpecializationInfo(currentSpec)) or "NONE"

    if event == "VARIABLES_LOADED" then
        if #RShamanStatArchive > 0 then
            DEFAULT_CHAT_FRAME:AddMessage("RSS: "..#RShamanStatArchive.." archived records found.");
        end
        RShamanStatArchive = rss.LoadArchive(RShamanStatArchive)
        return;
    end

	if role ~= "HEALER" then
		return;
    end

	if event == "PLAYER_REGEN_DISABLED" then
		-- entering combat; reset counters
		RShamanStatClear();
	end
	if event == "PLAYER_REGEN_ENABLED" then
		-- leaving combat: print status and reset counters
		RShamanStatReset();
	end
	if event == "COMBAT_LOG_EVENT_UNFILTERED" then
		RShamanStatCombatLogEntry(...);
	end
	if event == "UNIT_HEALTH_FREQUENT" or event == "UNIT_MAXHEALTH" then
		local unitID = select(1,...);
		if unitID == nil then return; end
		if UnitIsPlayer(unitID) or UnitPlayerOrPetInParty(unitID) or UnitPlayerOrPetInRaid(unitID) then
			local unitGUID = UnitGUID(unitID);
			if unitGUID == nil then return; end
			local unitHealthTable = healthTable[unitGUID];
			if unitHealthTable == nil then
				unitHealthTable = { };
				healthTable[unitGUID] = unitHealthTable;
			end
			unitHealthTable.maxHealth = UnitHealthMax(unitID);
			unitHealthTable.currentHealth = UnitHealth(unitID);
			if unitHealthTable.maxHealth < 1 then
				healthTable[unitGUID] = nil;
			end
		end
	end
end